# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import platform

from .framew.application import OperationError

##############################################################################


class Library (object):

    GENERIC_FALLBACK_MAVEN_BASE_URLS = ['https://libraries.minecraft.net/',
                                        'https://files.minecraftforge.net/maven/',
                                        'https://repo1.maven.org/maven2/']

    def __init__(self, library):
        super(Library, self).__init__()
        self.library = library
        parts = self.library['name'].split(':')
        self.group, self.artifact, self.version = parts[:3]
        self.type = parts[3] if len(parts) >= 4 else None
        self.extension = 'jar'
        if '@' in self.version:
            self.version, self.extension = self.version.split('@', 1)

        self.native = platform.system().lower()
        if self.native == 'darwin':
            self.native = 'osx'

        arch = platform.architecture()[0]
        if arch == '64bit':
            self. arch = '64'
        elif arch == '32bit':
            self.arch = '32'
        else:
            raise OperationError('Unknown architecure. Please report this! {}'.format(arch))

    def get_native_string(self):
        if 'natives' in self.library:
            ns = self.library['natives'][self.native]
            return ns.replace('${arch}', self.arch)
        return None

    def get_filename(self):
        nativestring = self.get_native_string()
        if nativestring:
            return '{}-{}-{}.{}'.format(self.artifact, self.version, nativestring, self.extension)
        elif self.type is not None:
            return '{}-{}-{}.{}'.format(self.artifact, self.version, self.type, self.extension)
        return '{}-{}.{}'.format(self.artifact, self.version, self.extension)

    def get_path(self):
        if 'path' in self.library:
            return self.library['path']

        if 'downloads' in self.library:
            nativestring = self.get_native_string()
            if nativestring:
                return self.library['downloads']['classifiers'][nativestring]['path']
            else:
                return self.library['downloads']['artifact']['path']

        return '{}/{}/{}/{}'.format(self.group.replace('.', '/'),
                                    self.artifact,
                                    self.version,
                                    self.get_filename())

    def get_path_local(self):
        path = self.get_path()
        return os.path.join(*path.split('/'))

    def get_download_url(self):
        if 'downloads' in self.library:
            nativestring = self.get_native_string()
            if nativestring:
                return self.library['downloads']['classifiers'][nativestring]['url'] or None
            else:
                return self.library['downloads']['artifact']['url'] or None
        if 'url' in self.library:
            return '{}{}'.format(self.library['url'], self.get_path())
        return None

    def get_download_urls(self):
        urls = []
        ref_url = self.get_download_url()
        if ref_url is not None:
            urls.append(ref_url)
        for urlbase in self.GENERIC_FALLBACK_MAVEN_BASE_URLS:
            urls.append('{}{}'.format(urlbase, self.get_path()))
        return urls

    def get_artifact(self):
        return '{}:{}'.format(self.group, self.artifact)

    def matches(self, environment):
        if 'rules' not in self.library:
            return True

        allow = False
        for rule in self.library['rules']:
            if self.matches_rule(rule, environment):
                allow = rule['action'] == 'allow'
        return allow

    def matches_rule(self, rule, environment):
        if 'os' in rule:
            return rule['os']['name'] == self.native
        return True

    def must_extract(self):
        return 'natives' in self.library

    def extract_excludes(self):
        excludes = []
        if 'extract' in self.library and 'exclude' in self.library['extract']:
            excludes = self.library['extract']['exclude']
        return excludes

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.resources as resources
import json
import requests

from packmaker.curse import api

##############################################################################


class MockApi (api.CurseAPI):

    def get_addon(self, projectId):
        try:
            resp = resources.read_text(self.__module__, 'get_addon_{}.json'.format(projectId))
            return json.loads(resp)
        except FileNotFoundError:
            raise requests.exceptions.HTTPError("not found")

    def get_addon_files(self, projectId):
        resp = resources.read_text(self.__module__, 'get_addon_files_{}.json'.format(projectId))
        return json.loads(resp)

    def get_addons_by_criteria(self, gameId, sectionId=None, categoryIds=None,
                               sort='Featured', sortDescending=True, gameVersions=None,
                               index=0, pageSize=1000, searchFilter=None):
        try:
            resp = resources.read_text(self.__module__, 'get_addons_by_criteria_{}.json'.format(searchFilter))
            return json.loads(resp)
        except FileNotFoundError:
            return []

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os.path
import requests
import shutil
import zipfile

from .curse.curseforge import Curseforge
from .framew.application import OperationError
from .framew.log import log_subprocess
from .libraries import Library
from .minecraft import Minecraft, version_as_integer
from .modloader import ModLoader

##############################################################################


class ForgeModLoader (ModLoader):

    FORGE_REPO_URL = 'https://files.minecraftforge.net/maven/'

    INCOMPATIBLE_FORGE_VERSIONS = ['14.23.5.2851']

    V19_BUILD_THAT_CHANGED_FORGE_VERSION_PATTERN = 1886
    V1122_BUILD_THAT_CHANGED_UNIVERSAL_FILENAME = 2851

    ##########################################################################

    def __init__(self, minecraft_version, loader_version, downloader):
        super(ForgeModLoader, self).__init__(minecraft_version, loader_version, downloader)
        self.installer_jar_filename = None
        self.install_profile = None
        self.version_info = None
        self.modloader_info = None

        if self.loader_version in ForgeModLoader.INCOMPATIBLE_FORGE_VERSIONS:
            raise OperationError('Sorry. Forge version {} does not work with '.format(self.forge_version)
                                 + 'packmaker. Upgrade to a newer version if possible.')

        self.calc_full_version()

    ##########################################################################

    def calc_full_version(self):
        mver = version_as_integer(self.minecraft_version)
        if mver < 10900:
            self.full_version = '{}-{}-{}'.format(self.minecraft_version, self.loader_version, self.minecraft_version)
        elif mver < 10904:
            build_num = int(self.loader_version.split('.')[-1])
            if build_num < ForgeModLoader.V19_BUILD_THAT_CHANGED_FORGE_VERSION_PATTERN:
                self.full_version = '{}-{}-{}'.format(self.minecraft_version, self.loader_version, self.minecraft_version)
            else:
                self.full_version = '{}-{}'.format(self.minecraft_version, self.loader_version)
        else:
            self.full_version = '{}-{}'.format(self.minecraft_version, self.loader_version)

    ##########################################################################

    def get_mainclass(self):
        version_info = self._get_version_info()
        return version_info['mainClass']

    ##########################################################################

    def get_launch_arguments(self):
        version_info = self._get_version_info()
        if 'arguments' in version_info:
            return ' '.join([arg for arg in version_info['arguments']['game'] if type(arg) == str])

        elif 'minecraftArguments' in version_info:
            return version_info['minecraftArguments']

    ##########################################################################

    def get_libraries(self, side=None):
        dependents = []
        version_info = self._get_version_info()
        libraries = version_info['libraries']
        for lib in libraries:
            library = Library(lib)
            if library.matches(None):
                dependents.append(library)
        return dependents

    ##########################################################################

    def install(self, install_location, side):
        self.log.info('Installing forge and dependent libraries ...')

        lib_location = os.path.join(install_location, 'libraries')
        if not os.path.exists(lib_location):
            os.makedirs(lib_location)

        self._install_libraries(lib_location, self.get_libraries())
        self._install_libraries(lib_location, self._get_installer_libraries())
        self._run_processors(lib_location, side)

        if side == 'server':
            if version_as_integer(self.minecraft_version) < 11400:
                self.log.info('Extract universal forge jar from installer...')
                universal_jar = os.path.join(install_location, self.get_server_jar_filename())
                self._get_universal_jar(universal_jar)
            else:
                src_jarfile = os.path.join(install_location, 'libraries',
                                           'net/minecraftforge/forge/{}/forge-{}.jar'.format(self.full_version,
                                                                                             self.full_version))
                dst_jarfile = os.path.join(install_location, self.get_server_jar_filename())
                shutil.copy2(src_jarfile, dst_jarfile)

    ##########################################################################

    def get_server_jar_filename(self):
        return 'forge_server-{}.jar'.format(self.full_version)

    ##########################################################################

    def _get_installer_jar(self):
        if self.installer_jar_filename is None:
            url = '{}net/minecraftforge/forge/{}/forge-{}-installer.jar'.format(ForgeModLoader.FORGE_REPO_URL,
                                                                                self.full_version,
                                                                                self.full_version)
            self.installer_jar_filename = self.downloader.download_immediate(url)
        return self.installer_jar_filename

    ##########################################################################

    def universal_download_url(self):
        return '{}net/minecraftforge/forge/{}/forge-{}-universal.jar'.format(ForgeModLoader.FORGE_REPO_URL,
                                                                             self.full_version,
                                                                             self.full_version)

    ##########################################################################

    def _universal_filename(self):
        mc_version = version_as_integer(self.minecraft_version)
        if mc_version < 11202:
            return 'forge-{}-universal.jar'.format(self.full_version)
        elif mc_version == 11202:
            build_num = int(self.loader_version.split('.')[-1])
            if build_num >= ForgeModLoader.V1122_BUILD_THAT_CHANGED_UNIVERSAL_FILENAME:
                return 'forge-{}.jar'.format(self.full_version)
        return 'forge-{}-universal.jar'.format(self.full_version)

    #########################################################################

    def _get_universal_jar(self, filename):
        installer_filename = self._get_installer_jar()
        with zipfile.ZipFile(installer_filename, 'r') as installerjar:
            try:
                universal_filename = installerjar.getinfo(self._universal_filename())
            except KeyError:
                try:
                    universal_filename = installerjar.getinfo('maven/net/minecraftforge/forge/{}/{}'.format(
                                                              self.full_version, self._universal_filename()))
                except KeyError:
                    raise Exception('Cannot find {} in the forge jar file: {}'
                                    .format(self._universal_filename(), installer_filename))

            with open(filename, 'wb') as outjar:
                with installerjar.open(universal_filename, 'r') as injar:
                    shutil.copyfileobj(injar, outjar)

    #########################################################################

    def _get_forge_info_from_curse(self):
        if self.modloader_info is None:
            api = Curseforge(None)  # no authn_token
            self.modloader_info = api.get_modloader_info('forge-{}'.format(self.loader_version))
        return self.modloader_info

    #########################################################################

    def _get_install_profile(self):
        # First, try to get the profile from the curse api ...
        if self.install_profile is None:
            profileString = self._get_forge_info_from_curse().get('installProfileJson', None)
            if profileString is not None:
                self.install_profile = json.loads(profileString)

        # If that doesn't work, download the installer jarfile and extract
        # the profile from there ...
        if self.install_profile is None:
            installer_filename = self._get_installer_jar()
            with zipfile.ZipFile(installer_filename, 'r') as installerjar:
                with installerjar.open('install_profile.json', 'r') as installf:
                    self.install_profile = json.load(installf)

        return self.install_profile

    ##########################################################################

    def _get_version_info(self):
        # First, try to get the version info the from curse api ...
        if self.version_info is None:
            versionString = self._get_forge_info_from_curse().get('versionJson', None)
            if versionString is not None:
                self.version_info = json.loads(versionString)

        # If that doesn't work, download the installer jarfile and
        # extract the profile from there ...
        if self.version_info is None:
            installer_filename = self._get_installer_jar()
            with zipfile.ZipFile(installer_filename, 'r') as installerjar:
                try:
                    with installerjar.open('version.json', 'r') as installf:
                        self.version_info = json.load(installf)
                except KeyError:
                    install_profile = self._get_install_profile()
                    if 'versionInfo' in install_profile:
                        self.version_info = install_profile['versionInfo']

        return self.version_info

    ############################################################################

    def _get_installer_libraries(self):
        dependents = []
        install_profile = self._get_install_profile()
        if 'libraries' in install_profile:
            for lib in install_profile['libraries']:
                library = Library(lib)
                if (library.artifact == "forge"
                   and library.type is not None
                   and library.type == "universal"
                   and library.get_download_url() is None):
                    # its okay to skip the forge universal library here if it doesn't
                    # have a download url in the manifest
                    continue
                if library.matches(None):
                    dependents.append(library)
        return dependents

    ############################################################################

    def _run_processors(self, lib_location, side):
        install_profile = self._get_install_profile()
        if 'processors' not in install_profile:
            return

        self.log.info('Processing installer data ...')
        data = {}
        if 'data' in install_profile:
            for datakey, dataval in install_profile['data'].items():
                dataval = dataval[side]
                if dataval.startswith('[') and dataval.endswith(']'):
                    # artifact
                    data[datakey] = os.path.join(lib_location,
                                                 _mavenname_to_path(dataval[1:-1]))

                elif dataval.startswith('\'') and dataval.endswith('\''):
                    # literal
                    data[datakey] = dataval[1:-1]

                else:
                    self.log.moreinfo('  Extracting: {}'.format(dataval))

                    data_location = os.path.join(lib_location, 'data')
                    if not os.path.exists(data_location):
                        os.makedirs(data_location)

                    installerjar = self._get_installer_jar()
                    with zipfile.ZipFile(installerjar, 'r') as installer:
                        installer.extract(dataval.lstrip('/'), data_location)

                    data[datakey] = os.path.join(data_location, dataval.lstrip('/'))

        manifest = Minecraft(self.minecraft_version).get_version_manifest()

        data['SIDE'] = side
        data['MINECRAFT_JAR'] = os.path.join(lib_location, 'com', 'mojang', 'minecraft',
                                             manifest['id'], 'minecraft-{}-{}.jar'.format(manifest['id'], side))

        self.log.info('Executing forge installer processors ...')
        processors = install_profile['processors']

        i = 0
        for proc in processors:
            i += 1
            self.log.moreinfo('  Running {} of {} processors ...'.format(i, len(processors)))

            classpath = proc['classpath']
            args = proc['args']

            finalargs = []
            for arg in args:
                if arg.startswith('{') and arg.endswith('}'):
                    arg = data[arg[1:-1]]
                elif arg.startswith('[') and arg.endswith(']'):
                    arg = os.path.join(lib_location,
                                       _mavenname_to_path(arg[1:-1]))

                finalargs.append(arg)

            classpath = [os.path.join(lib_location, _mavenname_to_path(proc['jar']))]
            for lib in proc['classpath']:
                classpath.append(os.path.join(lib_location, _mavenname_to_path(lib)))

            mainclass = _find_mainclass_in_jar(os.path.join(lib_location,
                                                            _mavenname_to_path(proc['jar'])))

            cmd = ['java', '-cp', ':'.join(classpath), mainclass]
            cmd.extend(finalargs)

            self.log.debug('cmd = {}'.format(' '.join(cmd)))

            returncode = log_subprocess(self.log.moreinfo, cmd)
            if returncode != 0:
                raise OperationError('Processor finished with errors. Abort')


##############################################################################

FORGE_PROMOTIONS_URL = "https://files.minecraftforge.net/maven/net/minecraftforge/forge/promotions_slim.json"


def calculate_forge_version(minecraft_version, forge_version):
    ver = '{}-{}'.format(minecraft_version, forge_version)
    promos = requests.get(FORGE_PROMOTIONS_URL, timeout=60).json()
    try:
        return promos['promos'][ver]
    except KeyError:
        return None

##############################################################################


def _mavenname_to_path(name):
    parts = name.split(':')
    filename = '{}-{}'.format(parts[2], parts[3]) if len(parts) >= 4 else parts[2]
    filename = filename.replace('@', '.') if '@' in filename else '{}.jar'.format(filename)
    initpath = parts[0].split('.')
    initpath.append(parts[1])
    initpath.append(parts[2].split('@')[0])
    initpath.append('{}-{}'.format(parts[1], filename))
    return '/'.join(initpath)


def _find_mainclass_in_jar(jarfile):
    with zipfile.ZipFile(jarfile, 'r') as jar:
        with jar.open('META-INF/MANIFEST.MF', 'r') as manifest:
            lines = manifest.readlines()
            for line in lines:
                line = line.decode('utf-8').strip()
                if line.startswith('Main-Class:'):
                    return line.split(' ')[1]
    return None

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aiohttp
import asyncio
import os
import urllib.parse

from typing import Any, Generator

from .framew.application import ApplicationError, OperationError
from .framew.log import getlog


##############################################################################

class DownloadError(OperationError):
    """Raised when a downloader fails to download a file"""
    def __init__(self, status):
        self.status = status


##############################################################################

class RetryableHttpClient(object):

    start_timeout = 0.1
    max_timeout = 30.0
    factor = 2.0

    def __init__(self, attempts=3, statuses=[413, 429, 500, 502, 504], exceptions=[], *args, **kwargs):
        self.attempts = attempts
        self.statuses = statuses
        self.exceptions = exceptions
        self.client = aiohttp.ClientSession(*args, **kwargs)
        self.closed = False

    def __del__(self):
        if not self.closed:
            print("!! RetryableHttpClient was not closed !!")

    def get(self, url, **kwargs):
        return self._request(self.client.get, url, **kwargs)

    def head(self, url, **kwargs):
        return self._request(self.client.head, url, **kwargs)

    def _request(self, request, url, **kwargs):
        return RetryableRequestContext(request,
                                       url,
                                       self.attempts,
                                       self.statuses,
                                       self.exceptions,
                                       self.retry_timeout,
                                       **kwargs)

    def retry_timeout(self, attempt):
        timeout = self.start_timeout * (self.factor ** attempt)
        return min(timeout, self.max_timeout)

    async def close(self):
        await self.client.close()
        self.closed = True

    async def __aenter__(self) -> 'RetryableHttpClient':
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()


class RetryableRequestContext(object):

    def __init__(self, request, url, attempts, retry_statuses, retry_exceptions, retry_timeout, **kwargs):
        self.request = request
        self.url = url
        self.attempts = attempts
        self.retry_statuses = retry_statuses
        self.retry_exceptions = retry_exceptions
        self.retry_timeout = retry_timeout
        self.kwargs = kwargs

        self.current_attempt = 0
        self.response = None

    async def do_request(self):
        try:
            self.current_attempt += 1
            response = await self.request(self.url, **self.kwargs)
            code = response.status
            if self.current_attempt < self.attempts and code in self.retry_statuses:
                retry_wait = self.retry_timeout(self.current_attempt)
                await asyncio.sleep(retry_wait)
                return await self.do_request()
            self.response = response
            return response
        except Exception as e:
            retry_wait = self.retry_timeout(self.current_attempt)
            if self.current_attempt < self.attempts:
                for exc in self.retry_exceptions:
                    if isinstance(e, exc):
                        await asyncio.sleep(retry_wait)
                        return await self.do_request()
            raise e

    def __await__(self) -> Generator[Any, None, aiohttp.ClientResponse]:
        return self.__aenter__().__await()

    async def __aenter__(self) -> aiohttp.ClientResponse:
        return await self.do_request()

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        if self.response is not None:
            if not self.response.closed:
                self.response.close()

##############################################################################


class HttpDownloader (object):

    MAX_CONCURRENT_DOWNLOADS = 5

    def __init__(self, download_dir):
        super(HttpDownloader, self).__init__()
        self.download_dir = download_dir
        self.log = getlog()
        self.download_list = []

    ##########################################################################

    def subdownloader(self, download_dir):
        return HttpDownloader(os.path.join(self.download_dir, download_dir))

    ##########################################################################

    def add_download(self, url, filename=None):
        """
        Add a url to download by this Downloader. The url can be either a plain
        string, an iterable of strings, or a callable that returns an iterable
        of strings.  If its an iterable, the downloader will try each one until
        one of the urls download successfully.

        Optionally specify the filename to write the downloaded url to.
        """
        if filename is None:
            filename = os.path.basename(urllib.parse.urlparse(url).path)
        localfile = os.path.join(self.download_dir, filename)
        self.download_list.append((url, localfile))
        return localfile

    ##########################################################################

    def run(self, post=None):
        asyncio.run(self.do_downloads(post))
        self.download_list = []

    ##########################################################################

    def download_immediate(self, url, filename=None, post=None):
        localfile = self.add_download(url, filename)
        self.run(post)
        return localfile

    ##########################################################################

    async def do_downloads(self, post):
        tasks = []
        sem = asyncio.Semaphore(self.MAX_CONCURRENT_DOWNLOADS)

        async with RetryableHttpClient() as client:
            for url, localfile in self.download_list:
                if not os.path.exists(localfile):
                    tasks.append(self.do_download_one(url, localfile, post, client, sem))
                else:
                    self.log.moreinfo('  [cached]    : {}'.format(url))
                    if post is not None:
                        post(url, localfile)

            return await asyncio.gather(*tasks)

    ##########################################################################

    async def do_download_one(self, url, dest, post, client, sem):
        async with sem:

            destdir = os.path.dirname(dest)
            if not os.path.exists(destdir):
                os.makedirs(destdir)
            tempfile = os.path.join(destdir, '.downloading_{}'.format(os.path.basename(dest)))
            if os.path.exists(tempfile):
                os.remove(tempfile)

            # convert the supplied url to an iterable of urls to try
            # the url could be:
            # - a iterable: nothing to do here
            # - a string:  "http://google.com" -> ["http://google.com"]
            # - a callable: [ u for u in url() ]
            if type(url) is str:
                urls = [url, ]
            elif callable(url):
                urls = [u for u in url()]
            else:
                try:
                    urls = iter(url)
                except TypeError:
                    raise ApplicationError("Invalid type passed as url: {}".format(url))

            try:
                for one_url in urls:
                    self.log.moreinfo('  downloading : {}'.format(one_url))
                    async with client.get(one_url) as resp:
                        content = await resp.read()
                    if resp.status == 200:
                        with open(tempfile, 'wb') as f:
                            f.write(content)
                        os.rename(tempfile, dest)
                        if post is not None:
                            post(one_url, dest)
                        break
                else:
                    raise DownloadError(status=resp.status)
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests

from ..framew.baseobject import BaseObject

##############################################################################


class CurseAPI (BaseObject):

    max_pagesize = 50
    max_index = 10000

    def __init__(self):
        super(BaseObject, self).__init__()

    def _api_baseurl(self):
        raise NotImplementedError

    def _request_headers(self, authRequired):
        return None

    def _getrequest(self, url, params=None, authRequired=True):
        r = requests.get(self._api_baseurl() + url, params, timeout=60, headers=self._request_headers(authRequired))
        r.raise_for_status()
        return r

    def _postrequest(self, url, data, authRequired=True):
        r = requests.post(self._api_baseurl() + url, json=data, timeout=60, headers=self._request_headers(authRequired))
        r.raise_for_status()
        return r

    def get_addon(self, projectId):
        return self._getrequest('addon/{}'.format(projectId)).json()

    def get_addon_file(self, projectId, fileId):
        return self._getrequest('addon/{}/file/{}'.format(projectId, fileId)).json()

    def get_addons(self, projectIds):
        return self._postrequest('addon', projectIds).json()

    def get_addon_files(self, projectId):
        return self._getrequest('addon/{}/files'.format(projectId)).json()

    def get_addons_by_criteria(self,
                               gameId,
                               sectionId=None,
                               categoryIds=None,
                               sort='Featured',
                               sortDescending=True,
                               gameVersion=None,
                               index=0,
                               pageSize=50,
                               searchFilter=None):
        url = 'addon/search'
        parameters = {"gameID": gameId,
                      "sort": sort,
                      "sortDescending": sortDescending,
                      "index": index,
                      "pageSize": pageSize
                      }

        if sectionId is not None:
            parameters['sectionId'] = sectionId
        if categoryIds is not None:
            parameters['categoryId'] = categoryIds
        if gameVersion is not None:
            parameters['gameVersion'] = gameVersion
        if searchFilter is not None:
            parameters['searchFilter'] = searchFilter

        return self._getrequest(url, parameters).json()

    def yield_addons_by_criteria(self,
                                 gameId,
                                 sectionId=None,
                                 categoryIds=None,
                                 sort='Featured',
                                 sortDescending=True,
                                 gameVersions=None,
                                 pageSize=50,
                                 searchFilter=None):

        if type(gameVersions) is not list:
            gameVersions = [gameVersions]

        if pageSize > self.max_pagesize:
            raise Exception("curse api limits pagesize to {}".format(self.max_pagesize))

        addons_seen = []
        for gameVersion in gameVersions:
            index = 0
            done = False
            while not done:
                results = self.get_addons_by_criteria(gameId, sectionId, categoryIds,
                                                      sort, sortDescending, gameVersion,
                                                      index, pageSize, searchFilter)
                index += pageSize
                if len(results) < pageSize:
                    done = True

                if index > self.max_index:
                    done = True
                else:
                    if index + pageSize > self.max_index:
                        pageSize = self.max_index - index
                        if pageSize < 1:
                            done = True

                for addon in results:
                    if addon['id'] not in addons_seen:
                        yield addon
                    addons_seen.append(addon['id'])

    def get_modloader_timestamp(self):
        return self._getrequest('minecraft/modloader/timestamp', authRequired=False).json()

    def get_modloader_list(self):
        return self._getrequest('minecraft/modloader', authRequired=False).json()

    def get_modloader_info(self, versionName):
        return self._getrequest('minecraft/modloader/{}'.format(versionName), authRequired=False).json()

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2019 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import dateutil.parser
import json
import os.path
import requests
import urllib.parse

from .curse.curseforge import Curseforge
from .framew.application import OperationError
from .framew.config import ConfigError
from .framew.log import getlog
from .packdef import ModDefinition, ResourcepackDefinition

##############################################################################

depend_types = ["_all_", "embeddedlib", "optional", "required", "tool", "incompatible", "include"]

release_types = ["_unknown", "release", "beta", "alpha"]


class Resolver (object):
    """
    Addon resolver logic, turning a simple addon definition from the packdef into a fully
    resolved addon object, with download urls and dependnecies resolved.
    """

    # A list of minecraft versions that are "fuzzily" equivalent, meaning that if a mod
    # specifies one version of minecraft as a dependency, the other versions should also
    # work correctly.  This helps when minecraft bumps a version up, but a version of the
    # mod is not released explicitly for the new minecraft version as there is no functional
    # code differences between them that would require a new mod release. A good example of
    # this is minecraft versions 1.16.4 and 1.16.5. Nearly all mods for the specify one of
    # these minecraft versions will also work on the other...
    fuzzy_mc_versions = [["1.16.4", "1.16.5"], ]

    def __init__(self, config, pack, fuzzy=True):
        super(Resolver, self).__init__()
        self.log = getlog()
        self.allow_fuzzy_mc_versions = fuzzy
        self.setup_api(config)
        self.setup_db(config)
        self.setup_pack(pack)

    def setup_api(self, config):
        authn_token = config.get('curseforge::authentication_token')
        if authn_token is None:
            raise ConfigError('No curseforge authentication token')
        self.api = Curseforge(authn_token)

    def setup_db(self, config):
        db_filename = config.get('curseforge::moddb_filename')
        if db_filename is None:
            self.log.warning('No moddb_filename parameter in configuration.'
                             ' This will force packmaker to make API calls to resolve every addon')
            self.db = None
            return

        self.log.moreinfo('Loading curseforge database ...')
        with open(db_filename, 'r') as cf:
            self.db = json.load(cf)

    def setup_pack(self, pack):
        self.mc_version = [pack.minecraft_version]
        if self.allow_fuzzy_mc_versions:
            for fuzzy in self.fuzzy_mc_versions:
                if pack.minecraft_version in fuzzy:
                    self.mc_version = fuzzy
                    break

        self.modloader = None
        if pack.forge_version is not None:
            self.modloader = 'forge'
        elif pack.fabric_version is not None:
            self.modloader = 'fabric'

        self.mod_releasetypes = pack.get_mod_releasetypes()
        self.resourcepack_releasetypes = pack.get_resourcepack_releasetypes()

        self.defined_mods = pack.get_all_mods()
        self.defined_resourcepacks = pack.get_all_resourcepacks()

    def resolve_addon(self, addon, addontype, packlock):
        self.log.moreinfo('Resolving: {}'.format(addon.slug))
        if addon.url is not None:
            # Find the addon by raw URL
            try:
                u = urllib.parse.urlparse(addon.url)
                filename = os.path.basename(u.path)
                packlock.add_resolved_addon(addon, addontype, {'fileName': filename, 'downloadUrl': addon.url})
            except ValueError:
                raise OperationError('Invalid addon url for {}'.format(addon.slug))

        else:
            # Find the addon in curseforge
            if addon.curseforgeid is not None:
                slug, name, addonid, author, website = self.find_addon_by_id(addon.curseforgeid, addontype)
            else:
                slug, name, addonid, author, website = self.find_addon_by_slug(addon.slug, addontype)
            found = self.find_addon_file(addonid, addon.version, addontype)
            if found is None:
                raise OperationError('Cannot find an addon file for {}'.format(addon.slug))
            packlock.add_resolved_addon(addon, addontype,
                                        {'projectId': addonid,
                                         'name': name,
                                         'author': author,
                                         'website': website,
                                         'fileId': found['id'],
                                         'fileName': found['fileName'],
                                         'version': found['displayName'],
                                         'downloadUrl': found['downloadUrl']
                                         })
            self.check_dependencies(found['dependencies'], addon.ignoredeps, addontype, packlock)

    def check_dependencies(self, dependencies, ignoredeps, addontype, packlock):
        self.log.moreinfo("Checking for dependencies...")
        if type(ignoredeps) is str and ignoredeps == "all":
            self.log.moreinfo('  all dependencies ignored ("ingoreddeps: all")')
            return
        if len(dependencies) == 0:
            self.log.moreinfo('  no dependencies')
            return

        for dep in dependencies:
            addonid = dep['addonId']
            deptype = dep['type']
            if deptype != 3:
                self.log.moreinfo('  ignoring dependency (id={}, type={} ({}))'.format(addonid, deptype, depend_types[deptype]))
                continue

            depslug, depname, depid, depauthors, depwebsite = self.find_addon_by_id(addonid, addontype)
            if depslug in ignoredeps:
                self.log.moreinfo('  Ignoring dependency: {}'.format(depslug))
                continue

            self.log.moreinfo('  Solving dependency: {}'.format(depslug))
            try:
                # search the packdef. is this addon already included?
                packlock.get_addon(depslug, addontype)
                self.log.moreinfo('  Dependent addon already resolved')
            except KeyError:
                # Not already included. Is it an already solved dependency?
                if addontype == 'mod':
                    definitions = self.defined_mods
                    addondef = ModDefinition(depslug, None)
                elif addontype == 'resourcepacks':
                    definitions = self.defined_resourcepacks
                    addondef = ResourcepackDefinition(depslug, None)
                for depdef in definitions:
                    if depdef.slug == depslug:
                        self.log.moreinfo('  Dependent addon already include in packdef')
                        break
                # Not included and not already solved: resolve this dependent addon
                else:
                    self.log.moreinfo('  Addon needs resolution: {}'.format(depslug))
                    self.resolve_addon(addondef, addontype, packlock)

    def find_addon_by_slug(self, slug, addontype):
        if self.db is not None and slug in self.db and self.db[slug]['category'] == addontype:
            return slug, \
                self.db[slug]['name'], \
                self.db[slug]['id'], \
                self.db[slug]['authors'][0]['name'], \
                self.db[slug]['websiteUrl']
        else:
            return self.manual_addon_search(slug, addontype)

    def find_addon_by_id(self, id, addontype):
        if self.db is not None:
            for slug, entry in self.db.items():
                if entry['id'] == id and entry['category'] == addontype:
                    return entry['slug'], \
                        entry['name'], \
                        id, \
                        entry['authors'][0]['name'], \
                        entry['websiteUrl']

        try:
            addon = self.api.get_addon(id)
        except requests.exceptions.HTTPError:
            raise OperationError('Cannot find an addon with curseforge id {}'.format(id))

        return addon['slug'], \
            addon['name'], \
            addon['id'], \
            addon['authors'][0]['name'], \
            addon['websiteUrl']

    def manual_addon_search(self, slug, addontype):
        self.log.warning('Cannot find addon in local db: {}. Manually searching...'.format(slug))

        if addontype == 'mod':
            sectionid = 6
        elif addontype == 'resourcepack':
            sectionid = 12
        else:
            raise Exception('unknown addontype passed to manual_addon_search()')

        searchresults = list(({'name': addon['name'], 'id': addon['id'], 'slug': addon['slug'],
                               'authors': addon['authors'], 'websiteUrl': addon['websiteUrl']}
                              for addon in self.api.yield_addons_by_criteria(gameId=432, sectionId=sectionid,
                                                                             gameVersions=self.mc_version,
                                                                             searchFilter=slug)))

        if len(searchresults) < 1:
            raise OperationError('Cannot find an addon named \'{}\''.format(slug))
        elif len(searchresults) > 1:
            self.log.warning('Multiple search results found ({}).  Looking for an exact match in results...'
                             .format(len(searchresults)))
            for sresult in searchresults:
                if sresult['slug'] == slug:
                    searchresult = sresult
                    self.log.warning('Found it! ... {} (id = {})'.format(searchresult['slug'], searchresult['id']))
                    break
            else:
                searchresult = searchresults[0]
                self.log.warning('No exact match found! Using the first one (this could be wildly wrong) ... {} (id = {})'
                                 .format(searchresult['slug'], searchresult['id']))
        else:
            searchresult = searchresults[0]
            self.log.warning('Found it! ... {} (id = {})'.format(searchresult['slug'], searchresult['id']))

        return (searchresult['slug'], searchresult['name'], searchresult['id'],
                searchresult['authors'][0]['name'], searchresult['websiteUrl'])

    def find_addon_file(self, id, required_version, addontype):
        latestTimestamp = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)
        addon_found = None

        # Is the required_version an integer? If so, we will use it as a curseforge file id
        try:
            rvint = int(required_version)
        except ValueError:
            rvint = None

        for addonfile in self.api.get_addon_files(id):
            if rvint is not None:
                # exact version specified as a curseforge id. Just use it, with no additional checks
                if addonfile['id'] == rvint:
                    return addonfile

            elif required_version not in ['latest', 'latest-alpha', 'latest-beta', 'latest-release']:
                # exact version specifed.  Just use it, with no additonal checks
                if addonfile['fileName'] == required_version or \
                   addonfile['displayName'] == required_version:
                    return addonfile

            else:
                for mc_ver in self.mc_version:
                    if mc_ver in addonfile['gameVersion']:
                        break
                else:
                    continue

                if self.modloader == 'forge' and 'Fabric' in addonfile['gameVersion']:
                    continue
                elif self.modloader == 'fabric' and 'Forge' in addonfile['gameVersion']:
                    continue

                if required_version == 'latest-release' and addonfile['releaseType'] > 1:
                    continue
                if required_version == 'latest-beta' and addonfile['releaseType'] > 2:
                    continue

                if required_version == 'latest':
                    if addontype == 'mod':
                        releasetypes = self.mod_releasetypes
                    elif addontype == 'resourcepack':
                        releasetypes = self.resourcepack_releasetypes
                    else:
                        releasetypes = None
                    if releasetypes is not None:
                        if addonfile['releaseType'] == 3 and 'alpha' not in releasetypes:
                            continue
                        if addonfile['releaseType'] == 2 and 'beta' not in releasetypes:
                            continue
                        if addonfile['releaseType'] == 1 and 'release' not in releasetypes:
                            continue

                timestamp = dateutil.parser.parse(addonfile['fileDate'])
                if timestamp > latestTimestamp:
                    addon_found = addonfile
                    latestTimestamp = timestamp

        return addon_found

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..framew.application import OperationError
from ..framew.cmdapplication import Subcommand
from ..framew.log import getlog
from ..fabric import calculate_fabric_version
from ..forge import calculate_forge_version
from ..packdef import PackDefinition
from ..resolver import Resolver

##############################################################################

depend_types = ["_all_", "embeddedlib", "optional", "required", "tool", "incompatible", "include"]


class LockCommand (Subcommand):
    """
    Lock the modpack. Find mod download urls, generate a packmaker.lock file.
    """

    name = 'lock'

    default_packmaker_yml = 'packmaker.yml'

    def get_cmdline_parser(self):
        parser = super(LockCommand, self).get_cmdline_parser()
        parser.add_argument('--strict', default=False, action='store_true',
                            help='enforce strict minecraft version matching')
        parser.add_argument('--no-strict', dest='strict', action='store_false',
                            help='do not enforce strict minecraft version matching [default]')
        parser.add_argument('packdef',
                            nargs='*',
                            default=[LockCommand.default_packmaker_yml],
                            help='modpack definition file')
        return parser

    def setup(self):
        super(LockCommand, self).setup()
        self.log = getlog()

    def run_command(self, parsed_args):
        self.log.info('Reading pack definition ...')
        pack = PackDefinition(parsed_args.packdef)
        pack.load()

        # warn if the main metadata fields are not defined in the packdef
        for m in ('name', 'title', 'version', 'authors', 'minecraft_version'):
            val = getattr(pack, m)
            if val is None or val == []:
                self.log.warn('No \'{}\' field defined in {}'.format(m, ', '.join(parsed_args.packdef)))

        packlock = pack.get_packlock()
        packlock.set_all_metadata(pack)

        # Calculate the Forge version
        if pack.forge_version:
            if pack.forge_version not in ['latest', 'recommended']:
                packlock.set_metadata('forge_version', pack.forge_version)
            else:
                self.log.info('Resolving forge version...')
                fversion = calculate_forge_version(pack.minecraft_version, pack.forge_version)
                if fversion is None:
                    raise OperationError('Sorry, Forge does not currently publish a "{}" version for Minecraft {}'
                                         .format(pack.forge_version, pack.minecraft_version))
                self.log.moreinfo('  {} = {}'.format(pack.forge_version, fversion))
                packlock.set_metadata('forge_version', fversion)

        # Calculate the Fabric version
        if pack.fabric_version:
            if pack.fabric_version not in ['latest', 'recommended']:
                packlock.set_metadata('fabric_version', pack.fabric_version)
            else:
                self.log.info('Resolving fabric version...')
                fversion = calculate_fabric_version(pack.minecraft_version, pack.fabric_version)
                self.log.moreinfo('  {} = {}'.format(pack.fabric_version, fversion))
                packlock.set_metadata('fabric_version', fversion)

        resolver = Resolver(self.config, pack, fuzzy=(not parsed_args.strict))

        self.log.info('Resolving mods...')
        for moddef in pack.get_all_mods():
            resolver.resolve_addon(moddef, 'mod', packlock)

        self.log.info('Resolving resourcepacks...')
        for rpdef in pack.get_all_resourcepacks():
            resolver.resolve_addon(rpdef, 'resourcepack', packlock)

        self.log.info('Adding files...')
        for filesdef in pack.files:
            packlock.add_files(filesdef)

        self.log.info('Adding extra options (if any)...')
        if pack.routhio is not None:
            packlock.add_extraopt('routhio', pack.routhio)

        self.log.info('Writing pack lock...')
        packlock.save()
        self.log.moreinfo('Done.')

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pprint
from requests.exceptions import HTTPError

from ..curse.curseforge import Curseforge
from ..framew.application import OperationError
from ..framew.cmdapplication import Subcommand
from ..framew.config import ConfigError

##############################################################################


class ApidebugCommand (Subcommand):
    """
    Debug curse api calls
    """

    name = 'apidebug'

    apidebug_cmds = ['get_addon',
                     'get_addons',
                     'get_addon_file',
                     'get_addon_files',
                     'get_addons_by_criteria',
                     'get_modloader_timestamp',
                     'get_modloader_list',
                     'get_modloader_info',
                     'yield_addons_by_criteria',
                     ]

    def get_cmdline_parser(self):
        parser = super(ApidebugCommand, self).get_cmdline_parser()
        parser.add_argument('command', choices=self.apidebug_cmds, help='api function')
        parser.add_argument('parameters', nargs='*', help='api function parameters')
        return parser

    def setup_command(self, _arguments):
        authn_token = self.config.get('curseforge::authentication_token')
        if authn_token is None:
            raise ConfigError('No curseforge authentication token')
        self.api = Curseforge(authn_token)

    def run_command(self, arguments):
        try:
            command = getattr(self.api, arguments.command)

            params = []
            kwparams = {}
            for param in arguments.parameters:
                if '=' in param:
                    k, v = [s.strip() for s in param.split('=', 1)]
                    if v.startswith('[') and v.endswith(']'):
                        kwparams[k] = [e.strip() for e in v[1:-1].split(',')]
                    else:
                        kwparams[k] = v
                else:
                    params.append(param)

            result = command(*params, **kwparams)
            pp = pprint.PrettyPrinter(indent=2)
            if arguments.command in ['yield_addons_by_criteria']:
                idx = 0
                for oneresult in result:
                    print("{}: ".format(idx))
                    pp.pprint(oneresult)
                    idx += 1
            else:
                pp.pprint(result)
        except TypeError as err:
            raise OperationError(err)
        except HTTPError as err:
            raise OperationError(err)

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import re
import requests
import zipfile

from .framew.application import OperationError
from .libraries import Library
from .modloader import ModLoader

##############################################################################


class FabricModLoader (ModLoader):

    FABRIC_META_URL = 'https://meta.fabricmc.net/v2/versions'

    #########################################################################

    def __init__(self, minecraft_version, loader_version, downloader):
        super(FabricModLoader, self).__init__(minecraft_version, loader_version, downloader)
        self.manifest = None

    #########################################################################

    def get_mainclass(self):
        manifest = self._get_fabric_manifest()
        return manifest['launcherMeta']['mainClass']['client']

    #########################################################################

    def get_launch_arguments(self):
        pass

    #########################################################################

    def get_libraries(self, side=None):
        dependents = []
        if side is None:
            side = 'client'
        manifest = self._get_fabric_manifest()

        for lib in ('loader', 'intermediary'):
            manifest_lib = manifest[lib]
            dependents.append(Library({'name': manifest_lib['maven'], 'url': 'https://maven.fabricmc.net/'}))

        libraries = manifest['launcherMeta']['libraries']
        for lib in libraries['common']:
            library = Library(lib)
            if library.matches(None):
                dependents.append(library)
        for lib in libraries[side]:
            library = Library(lib)
            if library.matches(None):
                dependents.append(library)
        return dependents

    #########################################################################

    def install(self, install_location, side):
        self.log.info('Installing fabric and dependent libraries ...')
        if side == 'client':
            self.install_client(install_location)
        elif side == 'server':
            self.install_server(install_location)

    def install_client(self, install_location):
        lib_location = os.path.join(install_location, 'libraries')
        if not os.path.exists(lib_location):
            os.makedirs(lib_location)
        self._install_libraries(lib_location, self.get_libraries('client'))

    def install_server(self, install_location):
        launcher_jar = os.path.join(install_location, 'fabric-server-launch.jar')
        with zipfile.ZipFile(launcher_jar, mode='w', compression=zipfile.ZIP_DEFLATED) as jarfile:

            with jarfile.open('META-INF/MANIFEST.MF', mode='w') as mf:
                mf.write('Manifest-Version: 1.0\n'.encode('utf-8'))
                mf.write('Main-Class: net.fabricmc.loader.launch.server.FabricServerLauncher\n'.encode('utf-8'))

            with jarfile.open('fabric-server-launch.properties', mode='w') as props:
                props.write('launch.mainClass=net.fabricmc.loader.launch.knot.KnotServer\n'.encode('utf-8'))

            services = {}
            for library in self.get_libraries('server'):
                downloaded_file = None
                for url in library.get_download_urls():
                    try:
                        downloaded_file = self.downloader.download_immediate(url, library.get_path())
                        break
                    except requests.exceptions.HTTPError as herr:
                        self.log.moreinfo('Failed downloading {} - {}'.format(url, herr))

                if downloaded_file is None:
                    raise OperationError('Cannot download library: {}'.format(library.get_filename()))

                with zipfile.ZipFile(downloaded_file, mode='r') as library_jar:
                    for entry in library_jar.infolist():
                        if entry.is_dir():
                            continue
                        name = entry.filename
                        if name.startswith('META-INF/services/'):
                            # service handling here
                            if name not in services:
                                services[name] = []

                            lines = []
                            with library_jar.open(name, mode='r') as serv:
                                lines = serv.readlines()

                            for line in [line.decode() for line in lines]:
                                line = re.sub('#.*$', '', line).strip()
                                if line and line not in services[name]:
                                    services[name].append(line)

                        elif name in jarfile.namelist():
                            # duplicate file. skipping ...
                            pass
                        else:
                            with library_jar.open(name, mode='r') as input:
                                with jarfile.open(name, mode='w') as output:
                                    output.write(input.read())

            # Write services
            for entry, content in services.items():
                with jarfile.open(entry, mode='w') as serv:
                    for line in content:
                        serv.write('{}\n'.format(line).encode('utf-8'))

        # fabric-server-launcher.properties file
        serverjar = os.path.join('libraries', 'com', 'mojang', 'minecraft', self.minecraft_version,
                                 'minecraft-{}-server.jar'.format(self.minecraft_version))
        with open(os.path.join(install_location, 'fabric-server-launcher.properties'), 'w') as props:
            props.write('serverJar={}\n'.format(serverjar))

    #########################################################################

    def get_server_jar_filename(self):
        return 'fabric-server-launch.jar'

    #########################################################################

    def _get_fabric_manifest(self):
        if self.manifest is None:
            manifest_url = self.FABRIC_META_URL + '/loader/{}/{}'.format(self.minecraft_version, self.loader_version)
            self.manifest = requests.get(manifest_url).json()
        return self.manifest

##############################################################################


FABRIC_LOADER_URL = "https://meta.fabricmc.net/v2/versions/loader/{}"


def calculate_fabric_version(minecraft_version, fabric_version):
    url = FABRIC_LOADER_URL.format(minecraft_version)
    loaders = requests.get(url, timeout=60).json()
    return loaders[0]['loader']['version']

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import shutil

from .framew.application import ApplicationError, OperationError
from .download import DownloadError
from .framew.log import getlog

##############################################################################


class ModLoader (object):

    def __init__(self, minecraft_version, loader_version, downloader):
        super(ModLoader, self).__init__()
        self.minecraft_version = minecraft_version
        self.loader_version = loader_version
        self.downloader = downloader
        self.log = getlog()

    ##########################################################################

    def get_mainclass(self):
        raise ApplicationError('get_mainclass() called base ModLoader class')

    ##########################################################################

    def get_launch_arguments(self):
        raise ApplicationError('get_launch_arguments() called base ModLoader class')

    ##########################################################################

    def get_libraries(self, side=None):
        raise ApplicationError('get_libraries() called base ModLoader class')

    ##########################################################################

    def install(self, install_location, side):
        raise ApplicationError('install() called base ModLoader class')

    ##########################################################################

    def get_server_jar_filename(self):
        raise ApplicationError('get_server_jar_filename() called base ModLoader class')

    ##########################################################################

    def _install_libraries(self, lib_location, libraries):

        def copy_library(url, filename):
            barepath = filename[len(self.downloader.download_dir):]
            local_library = os.path.join(lib_location, *barepath.split('/'))
            if not os.path.exists(os.path.dirname(local_library)):
                os.makedirs(os.path.dirname(local_library))
            shutil.copy2(filename, local_library)

        for library in libraries:
            self.downloader.add_download(library.get_download_urls, library.get_path())
        try:
            self.downloader.run(post=copy_library)
        except DownloadError:
            raise OperationError('Cannot download library: {}'.format(library.get_filename()))

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import json
import os
import requests
import shutil
import urllib.parse

from .base import BaseBuilder
from ..download import HttpDownloader
from ..framew.application import OperationError
from ..forge import ForgeModLoader
from ..minecraft import Minecraft, version_as_integer

##############################################################################


class RouthioBuilder (BaseBuilder):

    build_subloc = 'routhio'

    MAVEN_REPOS = ('https://libraries.minecraft.net/',
                   'https://repo1.maven.org/maven2/',
                   'https://files.minecraftforge.net/maven/'
                   )

    ##########################################################################

    def do_build(self):
        # Warn if building a pack for versions of Minecraft greater than 1.12.2
        if version_as_integer(self.packlock.get_metadata('minecraft_version')) > 11202:
            self.log.warning('!! The routhio launcher does not support modpacks for versions of Minecraft greater than 1.12.2')
            self.log.warning('!! Packmaker will build this pack, but the launcher may not be able to handle it.')

        manifest = {'title': self.packlock.get_metadata('title'),
                    'name': self.packlock.get_metadata('name'),
                    'version': self.packlock.get_metadata('version'),
                    'gameVersion': self.packlock.get_metadata('minecraft_version'),
                    'minimumVersion': 2,
                    'objectsLocation': 'objects',
                    'librariesLocation': 'libraries',
                    'tasks': []
                    }

        copytasks = self.copy_local_files(self.build_location('objects'), self.packlock.yield_clientonly_files())
        modtasks, features = self.process_mods()

        if features:
            manifest['features'] = features

        manifest['tasks'] = modtasks
        manifest['tasks'].extend(copytasks)

        version_manifest = self.get_minecraft_version_manifest()
        if self.packlock.get_metadata('forge_version') is not None:
            self.process_forge(version_manifest)
        manifest['versionManifest'] = version_manifest

        routhio = self.packlock.get_extraopt('routhio')
        if routhio is not None and 'launch' in routhio:
            manifest['launch'] = routhio['launch']

        self.log.info('Writing the pack manifest file...')
        with open(os.path.join(self.build_location(), '{}.json'.format(self.packlock.get_metadata('name'))), 'w') as mf:
            mf.write(json.dumps(manifest, indent=2, sort_keys=True))

        if routhio is not None and 'html' in routhio:
            self.log.info('Copy launcher html files ...')
            self.copy_files(files_iterator=routhio['html'])

        pkgfilename = os.path.join(self.release_location(),
                                   '{}-{}-routhio.{}'.format(self.packlock.get_metadata('name'),
                                                             self.packlock.get_metadata('version'),
                                                             self.release_extension))
        self.log.info('Packaging the modpack: {} ...'.format(pkgfilename))
        self.release_pkg(pkgfilename)

    ##########################################################################

    def copy_local_files(self, dest, files_iterator):
        self.log.info('Copying local files ...')
        tasks = []
        for fileloc in files_iterator:
            for root, dirs, files in os.walk(fileloc['location']):
                for file in files:
                    full_filename = os.path.join(root, file)

                    sha1hash = self.calc_sha1(full_filename)
                    sha1dir = os.path.join(sha1hash[0:2], sha1hash[2:4])

                    dest_directory = os.path.join(dest, sha1dir)
                    dest_filename = os.path.join(dest_directory, sha1hash)

                    if not os.path.exists(dest_directory):
                        os.makedirs(dest_directory)

                    shutil.copy2(full_filename, dest_filename)

                    to = os.path.join(root[len(fileloc['location']):], file)
                    location = os.path.join(sha1dir, sha1hash)

                    # The Launcher expects and requires all paths to use
                    # unix '/' path separators. So we need to ensure that any
                    # windoze-built paths are converted to unix paths here.

                    tasks.append({"type": 'file',
                                  "to": to.replace('\\', '/'),
                                  "size": os.path.getsize(full_filename),
                                  "location": location.replace('\\', '/'),
                                  "hash": sha1hash
                                  })
        return tasks

    ##########################################################################

    def process_mods(self):
        tasks = []
        features = []
        self.download_mods(self.packlock.yield_clientonly_mods(include_optional=True))
        for mod in self.packlock.yield_clientonly_mods(include_optional=True):
            localfile = os.path.join(self.cache_location('addons'), mod.fileName)
            if not os.path.exists(localfile):
                raise OperationError('Missing local copy of mod even after downloading: {}'.format(localfile))

            sha1hash = self.calc_sha1(localfile)
            task = {'type': 'url',
                    'to': 'mods/{}'.format(mod.fileName),
                    'location': mod.downloadUrl,
                    'hash': sha1hash,
                    'size': os.path.getsize(localfile)
                    }

            moddef = self.packlock.get_mod(mod.slug)
            if moddef.optional:
                task['when'] = {'features': [mod.name], 'if': 'requireAny'}
                feature = {'name': mod.name,
                           'description': moddef.description,
                           'recommendation': moddef.recommendation,
                           'selected': moddef.selected
                           }
                features.append(feature)

            tasks.append(task)
        return tasks, features

    ##########################################################################

    def calc_sha1(self, filename):
        h = hashlib.sha1()
        with open(filename, 'rb') as hf:
            data = hf.read(65536)
            while data:
                h.update(data)
                data = hf.read(65536)
        return h.hexdigest()

    ##########################################################################

    def get_minecraft_version_manifest(self):
        mc = Minecraft(self.packlock.get_metadata('minecraft_version'))
        version_manifest = mc.get_version_manifest()

        # tidy up a bit
        version_manifest.pop('logging', None)
        version_manifest.pop('downloads', None)

        return version_manifest

    ##########################################################################

    def process_forge(self, version_manifest):
        downloader = HttpDownloader(self.cache_location())
        forge = self.packlock.get_modloader(downloader)
        if not isinstance(forge, ForgeModLoader):
            raise OperationError('Routhio modpacks only work with the Forge modloader')

        self.log.info('Finding maven repositories for dependent libraries ...')
        downloader = downloader.subdownloader('forge')
        for library in forge.get_libraries():
            # cleanup the library entry a bit
            libdict = library.library
            libdict.pop('checksums', None)
            libdict.pop('serverreq', None)
            libdict.pop('clientreq', None)

            if libdict['name'].startswith('net.minecraftforge:forge:'):
                universal_url = forge.universal_download_url()
                universal_path = urllib.parse.urlparse(universal_url).path[1:]

                downloads = {'artifact': {'path': universal_path,
                                          'url': universal_url,
                                          }}
                libdict['downloads'] = downloads
                libdict.pop('url', None)
            else:
                downloaded_file = None
                for url in library.get_download_urls():
                    try:
                        downloaded_file = downloader.download_immediate(url, library.get_path())
                        break
                    except requests.exceptions.HTTPError as herr:
                        self.log.moreinfo('Failed downloading {} - {}'.format(url, herr))

                if downloaded_file is None:
                    raise OperationError('Cannot download library: {}'.format(library.get_filename()))

                path = library.get_path()
                libdict['url'] = url[:-len(path)]

            version_manifest['libraries'].append(libdict)

        for override in ('mainClass', 'minecraftArguments'):
            o = forge._get_version_info().get(override, None)
            if o is not None:
                version_manifest[override] = o

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import json
import os.path

from .base import BaseBuilder

try:
    import pkg_resources
    __version__ = pkg_resources.get_distribution('packmaker').version
except Exception:
    __version__ = 'unknown'

##############################################################################


class PackBuilder (BaseBuilder):

    build_subloc = 'pack'

    pack_version = '1'

    ##########################################################################

    def do_build(self):
        self.log.info('Generating metadata file ...')
        self.make_metadata()

        self.log.info('Generating manifest file ...')
        self.make_manifest()

        self.log.info('Copying local files ...')
        self.copy_files(self.build_location('files'), self.packlock.yield_clientonly_files())

        # add the packmaker "filemarker" file
        with open(os.path.join(self.build_location(), 'packmaker.json'), 'w') as pm:
            pm.write(json.dumps({'packmaker': __version__, 'format_version': self.pack_version}))

        pkgfilename = '{}/{}-{}.{}'.format(self.release_location(),
                                           self.packlock.get_metadata('name'),
                                           self.packlock.get_metadata('version'),
                                           'pack')
        self.log.info('Packaging the modpack: {} ...'.format(pkgfilename))
        self.release_pkg(pkgfilename, format='tbz2')

    def make_metadata(self):
        metadata = {}
        for k, v in self.packlock.get_all_metadata().items():
            if k not in ('minecraft_version', 'forge_version', 'fabric_version'):
                metadata[k] = v

        with open(os.path.join(self.build_location(), 'metadata.json'), 'w') as mf:
            mf.write(json.dumps(metadata))

    def make_manifest(self):

        #####

        def add_to_mod_manifest(url, filename):
            sha1hash = self.calc_sha1(filename)
            mod_manifest = {'type': 'url',
                            'destination': 'mods/{}'.format(filename),
                            'location': url,
                            'hash': sha1hash,
                            'size': os.path.getsize(filename)
                            }
            manifest['mods'].append(mod_manifest)

        #####

        def add_to_rp_manifest(url, filename):
            sha1hash = self.calc_sha1(filename)
            rp_manifest = {'type': 'url',
                           'destination': 'resourcepacks/{}'.format(filename),
                           'location': url,
                           'hash': sha1hash,
                           'size': os.path.getsize(filename)
                           }
            manifest['resourcepacks'].append(rp_manifest)

        #####

        manifest = {'minecraft': {'version': self.packlock.get_metadata('minecraft_version'),
                                  'modloaders': []
                                  },
                    'mods': [],
                    'resourcepacks': []
                    }

        forge_version = self.packlock.get_metadata('forge_version')
        if forge_version is not None:
            manifest['minecraft']['modloaders'].append({'id': 'forge', 'version': forge_version})

        fabric_version = self.packlock.get_metadata('fabric_version')
        if fabric_version is not None:
            manifest['minecraft']['modloaders'].append({'id': 'fabric', 'version': fabric_version})

        for mod in self.packlock.yield_clientonly_mods():
            self.addons_downloader.add_download(mod.downloadUrl, mod.fileName)
        self.addons_downloader.run(post=add_to_mod_manifest)

        for rp in self.packlock.get_all_resourcepacks():
            self.addons_downloader.add_download(rp.downloadUrl, rp.fileName)
        self.addons_downloader.run(post=add_to_rp_manifest)

        with open(os.path.join(self.build_location(), 'manifest.json'), 'w') as mf:
            mf.write(json.dumps(manifest))

    def calc_sha1(self, filename):
        h = hashlib.sha1()
        with open(filename, 'rb') as hf:
            data = hf.read(65536)
            while data:
                h.update(data)
                data = hf.read(65536)
        return h.hexdigest()


##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tarfile
import zipfile

from jinja2 import Template

from ..download import HttpDownloader
from ..framew.application import OperationError
from ..framew.log import getlog
from ..packlock import PackLock

##############################################################################


# A base class for builders that only performs the packlock loading
# functions. This is separated from the BaseBuilder so for the use
# of cmds the need to read the packlock files, but do not need all
# of the other building functions or parameters.
#
class BaseLockLoader (object):

    default_packmaker_lock = 'packmaker.lock'

    ##########################################################################

    def __init__(self):
        super(BaseLockLoader, self).__init__()
        self.log = getlog()

    ##########################################################################

    def add_cmdline_args(self, parser):
        parser.add_argument('lockfile', nargs='*', default=[BaseLockLoader.default_packmaker_lock],
                            help='modpack lock file')

    ##########################################################################

    def setup_packlocks(self, parsed_args):
        self.packlock_filenames = parsed_args.lockfile
        self.load_packfiles()

    ##########################################################################

    def load_packfiles(self):
        for filename in self.packlock_filenames:
            if not os.path.exists(filename):
                raise OperationError('Cannot find pack lock file: {}'.format(filename))

        self.packlock = PackLock(self.packlock_filenames)
        self.packlock.load()

##############################################################################


class BaseBuilder (BaseLockLoader):

    release_extensions = {'zip': 'zip',
                          'tgz': 'tar.gz',
                          'tbz2': 'tar.bz2',
                          'txz': 'tar.xz'
                          }

    # A sublocation that specific builders will use as their actual build_dir.
    # (for example, the local builder, specified build_subloc = 'local', and
    # then the builder will use the 'build/local' directory as its build location.
    # This allows multiple builders to work within the same build location with
    # out overwriting each others' content.
    build_subloc = None

    default_build_dir = 'build'

    ##########################################################################

    def __init__(self, config):
        super(BaseBuilder, self).__init__()
        self.config = config

    ##########################################################################

    def add_cmdline_args(self, parser):
        super(BaseBuilder, self).add_cmdline_args(parser)
        parser.add_argument('--build-dir', '-b', default=None,
                            help='base directory for build artifacts')
        parser.add_argument('--release-dir', '-r', default=None,
                            help='base directory for release artifacts')
        parser.add_argument('--cache-dir', default=None,
                            help='base directory for cached artifacts')
        parser.add_argument('--release-format', choices=('zip', 'tgz', 'tbz2', 'txz'), default='zip',
                            help='archive format for release package')

    ##########################################################################

    def setup_build(self, parsed_args):
        self.build_dir = parsed_args.build_dir
        if self.build_dir is None:
            self.build_dir = self.config.get('locations::build')
        if self.build_dir is None:
            self.build_dir = self.default_build_dir

        self.release_dir = parsed_args.release_dir
        if self.release_dir is None:
            self.release_dir = self.config.get('locations::release')
        if self.release_dir is None:
            self.release_dir = os.path.join(self.build_dir, 'release')

        self.cache_dir = parsed_args.cache_dir
        if self.cache_dir is None:
            self.cache_dir = self.config.get('locations::cache')
        if self.cache_dir is None:
            self.cache_dir = os.path.join(self.build_dir, 'cache')

        self.release_format = parsed_args.release_format
        self.release_extension = self.release_extensions[self.release_format]

        self.addons_downloader = HttpDownloader(self.cache_location('addons'))

        self.setup_packlocks(parsed_args)

    ##########################################################################

    def do_build(self):
        pass

    ##########################################################################

    def build_location(self, subloc=None):
        loc = self.build_dir
        if self.build_subloc is not None:
            loc = os.path.join(loc, self.build_subloc)
        if subloc is not None:
            loc = os.path.join(loc, subloc)
        if not os.path.exists(loc):
            os.makedirs(loc)
        return loc

    ##########################################################################

    def cache_location(self, subloc=None):
        loc = self.cache_dir
        if subloc is not None:
            loc = os.path.join(loc, subloc)
        if not os.path.exists(loc):
            os.makedirs(loc)
        return loc

    ##########################################################################

    def release_location(self, subloc=None):
        loc = self.release_dir
        if subloc is not None:
            loc = os.path.join(loc, subloc)
        if not os.path.exists(loc):
            os.makedirs(loc)
        return loc

    ##########################################################################

    def copy_files(self, dest=None, files_iterator=None):
        if dest is None:
            dest = self.build_location()
        if files_iterator is None:
            files_iterator = self.packlock.files
        for files in files_iterator:
            location = files['location']
            if not os.path.exists(location):
                self.log.warn('Skipping files, cannot find location: {}'.format(location))
                continue
            if not os.path.isdir(location):
                self.log.warn('Skipping files, location is not a directory: {}'.format(location))
                continue
            self.__copy_tree(files['location'], dest, files.get('template', False))

    ##########################################################################

    def copy_addon(self, dest, mod):
        def copy_file(url, temp):
            localfile = os.path.join(dest, os.path.basename(temp))
            shutil.copy2(temp, localfile)

        self.addons_downloader.add_download(mod.downloadUrl, mod.fileName)
        self.addons_downloader.run(post=copy_file)

    ##########################################################################

    def copy_mods(self, dest, mod_iterator=None):
        def copy_file(url, temp):
            localfile = os.path.join(dest, os.path.basename(temp))
            shutil.copy2(temp, localfile)

        if mod_iterator is None:
            mod_iterator = self.packlock.get_all_mods()
        for mod in mod_iterator:
            self.addons_downloader.add_download(mod.downloadUrl, mod.fileName)
        self.addons_downloader.run(post=copy_file)

    ##########################################################################

    def copy_resourcepacks(self, dest, rep_iterator=None):
        def copy_file(url, temp):
            localfile = os.path.join(dest, os.path.basename(temp))
            shutil.copy2(temp, localfile)

        if rep_iterator is None:
            rep_iterator = self.packlock.get_all_resourcepacks()
        for rep in rep_iterator:
            self.addons_downloader.add_download(rep.downloadUrl, rep.fileName)
        self.addons_downloader.run(post=copy_file)

    ##########################################################################

    def download_mods(self, mod_iterator=None):
        self.log.debug('Downloading mods ...')
        if mod_iterator is None:
            mod_iterator = self.packlock.get_all_mods()
        for mod in mod_iterator:
            self.addons_downloader.add_download(mod.downloadUrl, mod.fileName)
        self.addons_downloader.run()

    ##########################################################################

    def download_resourcepacks(self, rep_iterator=None):
        self.log.debug('Downloading resourcepacks ...')
        if rep_iterator is None:
            rep_iterator = self.packlock.get_all_resourcepacks()
        for rep in rep_iterator:
            self.addons_downloader.add_download(rep.downloadUrl, rep.fileName)
        self.addons_downloader.run()

    ##########################################################################

    def release_pkg(self, dest, source=None, format=None):
        if source is None:
            source = self.build_location()
        if format is None:
            format = self.release_format
        if format == 'zip':
            self.__zip_tree(source, dest)
        elif format == 'tgz':
            self.__tar_tree(source, dest, compression='gz')
        elif format == 'tbz2':
            self.__tar_tree(source, dest, compression='bz2')
        elif format == 'txz':
            self.__tar_tree(source, dest, compression='xz')
        else:
            raise OperationError('Unknown release format: {}'.format(self.release_format))

    ##########################################################################

    def __copy_tree(self, source, dest, template=False):
        names = os.listdir(source)
        if not os.path.exists(dest):
            os.makedirs(dest)
        for name in names:
            srcname = os.path.join(source, name)
            dstname = os.path.join(dest, name)
            try:
                if os.path.isdir(srcname):
                    self.__copy_tree(srcname, dstname, template)
                else:
                    if template:
                        self.__copy_template(srcname, dstname)
                    else:
                        shutil.copy2(srcname, dstname)
            except OSError:
                raise

    def __copy_template(self, srcname, dstname):
        with open(srcname, 'r') as src:
            templ = Template(src.read())
            out = templ.render(self.packlock.get_all_metadata())
            with open(dstname, 'w') as dst:
                dst.write(out)

    ##########################################################################

    @staticmethod
    def __zip_tree(source, dest):
        with zipfile.ZipFile(dest, 'w', zipfile.ZIP_DEFLATED) as zipf:
            for root, dirs, files in os.walk(source):
                for file in files:
                    # bad attempt at preventing the zipfile from including itself
                    if file == os.path.basename(dest):
                        continue
                    zipf.write(os.path.join(root, file),
                               os.path.join(root[len(source):], file))

    @staticmethod
    def __tar_tree(source, dest, compression='gz'):
        with tarfile.open(dest, 'w:{}'.format(compression)) as tarf:
            for root, dirs, files in os.walk(source):
                for file in files:
                    # bad attempt at preventing the tarfile from including itself
                    if file == os.path.basename(dest):
                        continue
                    tarf.add(os.path.join(root, file),
                             os.path.join(root[len(source):], file))

##############################################################################
# THE END



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..builder.base import BaseLockLoader
from ..framew.cmdapplication import Subcommand
from ..framew.outputter import SingleOutputter
from ..fabric import FabricModLoader
from ..forge import ForgeModLoader

##############################################################################


class InfoCommand (Subcommand):
    """
    Display information about the modpack.
    """

    name = 'info'

    def setup(self):
        super(InfoCommand, self).setup()
        self.lockloader = BaseLockLoader()
        self.outputter = SingleOutputter()

    def get_cmdline_parser(self):
        parser = super(InfoCommand, self).get_cmdline_parser()
        self.lockloader.add_cmdline_args(parser)
        self.outputter.add_argument_group(parser)
        return parser

    def run_command(self, arguments):
        self.lockloader.setup_packlocks(arguments)
        packlock = self.lockloader.packlock

        keys = []
        vals = []

        self.calc_metadata(packlock, keys, vals)
        self.calc_modloader(packlock, keys, vals)

        self.outputter.produce_output(arguments, keys, vals)

    def calc_metadata(self, packlock, keys, vals):
        metadata = packlock.get_all_metadata()
        keys.extend([k for k in metadata.keys()])
        vals.extend([v for v in metadata.values()])

    def calc_modloader(self, packlock, keys, vals):
        modloader = packlock.get_modloader(None)
        keys.append('modloader_jarfile')
        if modloader is not None:
            vals.append(modloader.get_server_jar_filename())
        else:
            vals.append("")

        keys.append('forge_jarfile')
        if isinstance(modloader, ForgeModLoader):
            vals.append(modloader.get_server_jar_filename())
        else:
            vals.append("")

        keys.append('fabric_jarfile')
        if isinstance(modloader, FabricModLoader):
            vals.append(modloader.get_server_jar_filename())
        else:
            vals.append("")



# vim:set ts=4 sw=4 et nowrap syntax=python ff=unix:
#
# Copyright 2020 Mark Crewson <mark@crewson.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .framew.cmdapplication import CommandApplication
from .cmd import commands

try:
    import pkg_resources
    __version__ = pkg_resources.get_distribution('packmaker').version
except Exception:
    __version__ = 'unknown'

##############################################################################


class PackmakerMain (CommandApplication):
    app_name = 'packmaker'
    app_version = __version__
    subcommands = commands
    censored_config_params = ['curseforge::authentication_token']

##############################################################################


def main():
    app = PackmakerMain()
    return app.start()


if __name__ == "__main__":
    main()

##############################################################################
# THE END
